# Standard library imports
import os
import threading
import time
import zipfile
from collections.abc import Callable
from typing import Any
from concurrent.futures import ThreadPoolExecutor

# Third party imports
import flask
import fastjsonschema  # type: ignore
import importlib.metadata as metadata
import shutil
from werkzeug.exceptions import HTTPException
import werkzeug

# Local application imports
from . import geode_functions
from opengeodeweb_microservice.database.data import Data
from opengeodeweb_microservice.database.connection import get_session


def increment_request_counter(current_app: flask.Flask) -> None:
    if "REQUEST_COUNTER" in current_app.config:
        REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER", 0))
        REQUEST_COUNTER += 1
        current_app.config.update(REQUEST_COUNTER=REQUEST_COUNTER)


def decrement_request_counter(current_app: flask.Flask) -> None:
    if "REQUEST_COUNTER" in current_app.config:
        REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER", 0))
        REQUEST_COUNTER -= 1
        current_app.config.update(REQUEST_COUNTER=REQUEST_COUNTER)


def update_last_request_time(current_app: flask.Flask) -> None:
    if "LAST_REQUEST_TIME" in current_app.config:
        LAST_REQUEST_TIME = time.time()
        current_app.config.update(LAST_REQUEST_TIME=LAST_REQUEST_TIME)


def before_request(current_app: flask.Flask) -> None:
    increment_request_counter(current_app)


def teardown_request(current_app: flask.Flask) -> None:
    decrement_request_counter(current_app)
    update_last_request_time(current_app)


def kill_task(current_app: flask.Flask) -> None:
    REQUEST_COUNTER = int(current_app.config.get("REQUEST_COUNTER", 0))
    LAST_PING_TIME = float(current_app.config.get("LAST_PING_TIME", 0))
    LAST_REQUEST_TIME = float(current_app.config.get("LAST_REQUEST_TIME", 0))
    MINUTES_BEFORE_TIMEOUT = float(current_app.config.get("MINUTES_BEFORE_TIMEOUT", 0))
    current_time = time.time()
    minutes_since_last_request = (current_time - LAST_REQUEST_TIME) / 60
    minutes_since_last_ping = (current_time - LAST_PING_TIME) / 60

    if REQUEST_COUNTER > 0:
        return
    if MINUTES_BEFORE_TIMEOUT == 0:
        return
    if minutes_since_last_ping > MINUTES_BEFORE_TIMEOUT:
        kill_server()
    if minutes_since_last_request > MINUTES_BEFORE_TIMEOUT:
        kill_server()


def kill_server() -> None:
    print("Server timed out due to inactivity, shutting down...", flush=True)
    os._exit(0)


def versions(list_packages: list[str]) -> list[dict[str, str]]:
    list_with_versions = []
    for package in list_packages:
        list_with_versions.append(
            {"package": package, "version": metadata.distribution(package).version}
        )
    return list_with_versions


def validate_request(request: flask.Request, schema: dict[str, str]) -> None:
    json_data = request.get_json(force=True, silent=True)

    if json_data is None:
        json_data = {}

    try:
        validate = fastjsonschema.compile(schema)
        validate(json_data)
    except fastjsonschema.JsonSchemaException as e:
        error_msg = str(e)
        flask.abort(400, error_msg)


def set_interval(
    function: Callable[[Any], None], seconds: float, args: Any
) -> threading.Timer:
    def function_wrapper() -> None:
        set_interval(function, seconds, args)
        function(args)

    timer = threading.Timer(seconds, function_wrapper)
    timer.daemon = True
    timer.start()
    return timer


def extension_from_filename(filename: str) -> str:
    return os.path.splitext(filename)[1][1:]


def send_file(
    upload_folder: str, saved_files: str, new_file_name: str
) -> flask.Response:
    if len(saved_files) == 1:
        mimetype = "application/octet-binary"
    else:
        mimetype = "application/zip"
        new_file_name = os.path.splitext(new_file_name)[0] + ".zip"
        with zipfile.ZipFile(os.path.join(upload_folder, new_file_name), "w") as zipObj:
            for saved_file_path in saved_files:
                zipObj.write(
                    saved_file_path,
                    os.path.basename(saved_file_path),
                )

    response = flask.send_from_directory(
        directory=upload_folder,
        path=new_file_name,
        as_attachment=True,
        mimetype=mimetype,
    )
    response.headers["new-file-name"] = new_file_name
    response.headers["Access-Control-Expose-Headers"] = "new-file-name"

    return response


def handle_exception(exception: HTTPException) -> flask.Response:
    response = exception.get_response()
    response.data = flask.json.dumps(
        {
            "code": exception.code,
            "name": exception.name,
            "description": exception.description,
        }
    )
    response.content_type = "application/json"
    return response


def create_data_folder_from_id(data_id: str) -> str:
    base_data_folder = flask.current_app.config["DATA_FOLDER_PATH"]
    data_path = os.path.join(base_data_folder, data_id)
    os.makedirs(data_path, exist_ok=True)
    return data_path


def save_all_viewables_and_return_info(
    geode_object: str,
    data: Any,
    input_file: str,
    additional_files: list[str] | None = None,
) -> dict[str, Any]:
    if additional_files is None:
        additional_files = []

    data_entry = Data.create(
        geode_object=geode_object,
        input_file=input_file,
        additional_files=additional_files,
    )
    data_path = create_data_folder_from_id(data_entry.id)
    with ThreadPoolExecutor() as executor:
        native_future = executor.submit(
            geode_functions.save,
            geode_object,
            data,
            data_path,
            "native." + data.native_extension(),
        )
        viewable_future = executor.submit(
            geode_functions.save_viewable, geode_object, data, data_path, "viewable"
        )
        light_viewable_future = executor.submit(
            geode_functions.save_light_viewable,
            geode_object,
            data,
            data_path,
            "light_viewable",
        )
        saved_light_viewable_file_path = light_viewable_future.result()
        with open(saved_light_viewable_file_path, "rb") as f:
            binary_light_viewable = f.read()
        saved_native_file_path = native_future.result()
        saved_viewable_file_path = viewable_future.result()
        data_entry.native_file_name = os.path.basename(saved_native_file_path[0])
        data_entry.viewable_file_name = os.path.basename(saved_viewable_file_path)
        data_entry.light_viewable = os.path.basename(saved_light_viewable_file_path)

        session = get_session()
        if session:
            session.commit()

        return {
            "native_file_name": data_entry.native_file_name,
            "viewable_file_name": data_entry.viewable_file_name,
            "id": data_entry.id,
            "object_type": geode_functions.get_object_type(geode_object),
            "binary_light_viewable": binary_light_viewable.decode("utf-8"),
            "geode_object": data_entry.geode_object,
            "input_files": data_entry.input_file,
            "additional_files": data_entry.additional_files,
        }


def generate_native_viewable_and_light_viewable_from_object(
    geode_object: str, data: Any
) -> dict[str, Any]:
    return save_all_viewables_and_return_info(geode_object, data, input_file="")


def generate_native_viewable_and_light_viewable_from_file(
    geode_object: str, input_filename: str
) -> dict[str, Any]:
    temp_data_entry = Data.create(
        geode_object=geode_object,
        input_file=input_filename,
        additional_files=[],
    )

    data_path = create_data_folder_from_id(temp_data_entry.id)

    full_input_filename = geode_functions.upload_file_path(input_filename)
    copied_full_path = os.path.join(
        data_path, werkzeug.utils.secure_filename(input_filename)
    )
    shutil.copy2(full_input_filename, copied_full_path)

    additional_files_copied: list[str] = []
    additional = geode_functions.additional_files(geode_object, full_input_filename)
    for additional_file in additional.mandatory_files + additional.optional_files:
        if additional_file.is_missing:
            continue
        source_path = os.path.join(
            os.path.dirname(full_input_filename), additional_file.filename
        )
        if not os.path.exists(source_path):
            continue
        dest_path = os.path.join(data_path, additional_file.filename)
        os.makedirs(os.path.dirname(dest_path), exist_ok=True)
        shutil.copy2(source_path, dest_path)
        additional_files_copied.append(additional_file.filename)

    data = geode_functions.load(geode_object, copied_full_path)

    # Remplacer :
    # database.session.delete(temp_data_entry)
    # database.session.flush()
    # Par :
    session = get_session()
    if session:
        session.delete(temp_data_entry)
        session.flush()

    return save_all_viewables_and_return_info(
        geode_object,
        data,
        input_file=input_filename,
        additional_files=additional_files_copied,
    )



# Standard library imports
import re
import os

# Third party imports
import flask
import shutil
import uuid

# Local application imports
from opengeodeweb_microservice.database.data import Data
from opengeodeweb_microservice.database.connection import get_session
from src.opengeodeweb_back import geode_functions, utils_functions


def test_increment_request_counter(app_context):
    assert flask.current_app.config.get("REQUEST_COUNTER") == 0
    utils_functions.increment_request_counter(flask.current_app)
    assert flask.current_app.config.get("REQUEST_COUNTER") == 1


def test_decrement_request_counter(app_context):
    assert flask.current_app.config.get("REQUEST_COUNTER") == 1
    utils_functions.decrement_request_counter(flask.current_app)
    assert flask.current_app.config.get("REQUEST_COUNTER") == 0


def test_update_last_request_time(app_context):
    LAST_REQUEST_TIME = flask.current_app.config.get("LAST_REQUEST_TIME")
    utils_functions.update_last_request_time(flask.current_app)
    assert flask.current_app.config.get("LAST_REQUEST_TIME") >= LAST_REQUEST_TIME


def test_before_request(app_context):
    assert flask.current_app.config.get("REQUEST_COUNTER") == 0
    utils_functions.before_request(flask.current_app)
    assert flask.current_app.config.get("REQUEST_COUNTER") == 1


def test_teardown_request(app_context):
    LAST_REQUEST_TIME = flask.current_app.config.get("LAST_REQUEST_TIME")
    assert flask.current_app.config.get("REQUEST_COUNTER") == 1
    utils_functions.teardown_request(flask.current_app)
    assert flask.current_app.config.get("REQUEST_COUNTER") == 0
    assert flask.current_app.config.get("LAST_REQUEST_TIME") >= LAST_REQUEST_TIME


def test_versions():
    list_packages = [
        "OpenGeode-core",
        "OpenGeode-IO",
        "OpenGeode-Geosciences",
        "OpenGeode-GeosciencesIO",
    ]
    versions = utils_functions.versions(list_packages)
    assert type(versions) is list
    for version in versions:
        assert type(version) is dict


def test_extension_from_filename():
    extension = utils_functions.extension_from_filename("test.toto")
    assert type(extension) is str
    assert extension.count(".") == 0


def test_handle_exception(client):
    route = "/error"
    response = client.post(route)
    assert response.status_code == 500
    data = response.get_json()
    assert type(data) is dict
    assert type(data["description"]) is str
    assert type(data["name"]) is str
    assert type(data["code"]) is int


def test_create_data_folder_from_id(client):
    app = client.application
    with app.app_context():
        test_id = str(uuid.uuid4()).replace("-", "")
        data_path = utils_functions.create_data_folder_from_id(test_id)
        assert isinstance(data_path, str)
        assert os.path.exists(data_path)
        assert data_path.startswith(flask.current_app.config["DATA_FOLDER_PATH"])
        assert test_id in data_path
        shutil.rmtree(data_path, ignore_errors=True)
        assert not os.path.exists(data_path)


def test_save_all_viewables_and_return_info(client):
    app = client.application
    with app.app_context():
        base_dir = os.path.abspath(os.path.dirname(__file__))
        expected_db_path = os.path.join(base_dir, "data", "project.db")
        expected_uri = f"sqlite:///{expected_db_path}"

        assert app.config["SQLALCHEMY_DATABASE_URI"] == expected_uri
        assert os.path.exists(expected_db_path)

        geode_object = "BRep"
        data = geode_functions.load(geode_object, "./tests/data/test.og_brep")
        input_file = "test.og_brep"
        additional_files = ["additional_file.txt"]

        result = utils_functions.save_all_viewables_and_return_info(
            geode_object, data, input_file, additional_files
        )

        assert isinstance(result, dict)
        assert result["native_file_name"].startswith("native.")
        assert result["viewable_file_name"].endswith(".vtm")
        assert isinstance(result["id"], str)
        assert len(result["id"]) == 32
        assert re.match(r"[0-9a-f]{32}", result["id"])
        assert isinstance(result["object_type"], str)
        assert isinstance(result["binary_light_viewable"], str)
        assert result["geode_object"] == geode_object
        assert result["input_files"] == input_file

        db_entry = Data.get(result["id"])
        assert db_entry is not None
        assert db_entry.native_file_name == result["native_file_name"]
        assert db_entry.viewable_file_name == result["viewable_file_name"]
        assert db_entry.geode_object == geode_object
        assert db_entry.input_file == input_file
        assert db_entry.additional_files == additional_files

        expected_data_path = os.path.join(app.config["DATA_FOLDER_PATH"], result["id"])
        assert os.path.exists(expected_data_path)


def test_save_all_viewables_commits_to_db(client):
    app = client.application
    with app.app_context():
        geode_object = "BRep"
        data = geode_functions.load(geode_object, "./tests/data/test.og_brep")
        input_file = "test.og_brep"
        result = utils_functions.save_all_viewables_and_return_info(
            geode_object, data, input_file
        )
        data_id = result["id"]
        db_entry_before = Data.get(data_id)
        assert db_entry_before is not None
        assert db_entry_before.native_file_name == result["native_file_name"]
        session = get_session()
        session.rollback()
        db_entry_after = Data.get(data_id)
        assert (
            db_entry_after is not None
        ), "database.session.commit() was not called - entry missing after rollback"
        assert db_entry_after.native_file_name == result["native_file_name"]


def test_generate_native_viewable_and_light_viewable_from_object(client):
    app = client.application
    with app.app_context():
        geode_object = "BRep"
        data = geode_functions.load(geode_object, "./tests/data/test.og_brep")

        result = (
            utils_functions.generate_native_viewable_and_light_viewable_from_object(
                geode_object, data
            )
        )

    assert isinstance(result, dict)
    assert isinstance(result["native_file_name"], str)
    assert result["native_file_name"].startswith("native.")
    assert isinstance(result["viewable_file_name"], str)
    assert result["viewable_file_name"].endswith(".vtm")
    assert isinstance(result["id"], str)
    assert re.match(r"[0-9a-f]{32}", result["id"])
    assert isinstance(result["object_type"], str)
    assert isinstance(result["binary_light_viewable"], str)
    assert result["input_files"] == ""


def test_generate_native_viewable_and_light_viewable_from_file(client):
    app = client.application
    with app.app_context():
        geode_object = "BRep"
        input_filename = "test.og_brep"

        result = utils_functions.generate_native_viewable_and_light_viewable_from_file(
            geode_object, input_filename
        )

    assert isinstance(result, dict)
    assert isinstance(result["native_file_name"], str)
    assert result["native_file_name"].startswith("native.")
    assert isinstance(result["viewable_file_name"], str)
    assert result["viewable_file_name"].endswith(".vtm")
    assert isinstance(result["id"], str)
    assert re.match(r"[0-9a-f]{32}", result["id"])
    assert isinstance(result["object_type"], str)
    assert isinstance(result["binary_light_viewable"], str)
    assert isinstance(result["input_files"], str)



import json
import os
import xml.etree.ElementTree as ET
import flask
from ... import geode_functions, utils_functions

routes = flask.Blueprint("models", __name__, url_prefix="/models")


schemas = os.path.join(os.path.dirname(__file__), "schemas")

with open(os.path.join(schemas, "vtm_component_indices.json"), "r") as file:
    vtm_component_indices_json = json.load(file)


@routes.route(
    vtm_component_indices_json["route"], methods=vtm_component_indices_json["methods"]
)
def uuid_to_flat_index():
    utils_functions.validate_request(flask.request, vtm_component_indices_json)

    vtm_file_path = geode_functions.data_file_path(
        flask.request.get_json().get("id"), "viewable.vtm"
    )
    tree = ET.parse(vtm_file_path)
    root = tree.find("vtkMultiBlockDataSet")
    uuid_to_flat_index = {}
    current_index = 0
    for elem in root.iter():
        if "uuid" in elem.attrib and elem.tag == "DataSet":
            uuid_to_flat_index[elem.attrib["uuid"]] = current_index
        current_index += 1
    return flask.make_response({"uuid_to_flat_index": uuid_to_flat_index}, 200)


def extract_model_uuids(model):
    mesh_components = model.mesh_components()
    uuid_dict = {}
    for mesh_component, ids in mesh_components.items():
        component_name = mesh_component.get()
        uuid_dict[component_name] = [id.string() for id in ids]
    return uuid_dict


with open(os.path.join(schemas, "mesh_components.json"), "r") as file:
    mesh_components_json = json.load(file)


@routes.route(mesh_components_json["route"], methods=mesh_components_json["methods"])
def extract_uuids_endpoint():
    utils_functions.validate_request(flask.request, mesh_components_json)
    model = geode_functions.load_data(flask.request.get_json().get("id"))
    uuid_dict = extract_model_uuids(model)
    return flask.make_response({"uuid_dict": uuid_dict}, 200)
